#include "object_recognition_header.hpp"
#include "globals.hpp"
using namespace cv;
using namespace std;

//Checks if a given point pt is inside the Quadrilateral given by corners.
bool point_is_inside(vector<Point2f> corners, Point2f pt)
{
  float x1, x2, y1, y2;
  x1=((pt.y-corners[0].y)*(corners[3].x-corners[0].x))/(corners[3].y-corners[0].y) + corners[0].x;
  x2=((pt.y-corners[2].y)*(corners[2].x-corners[1].x))/(corners[2].y-corners[1].y) + corners[2].x;
  y1=((pt.x-corners[0].x)*(corners[1].y-corners[0].y))/(corners[1].x-corners[0].x) + corners[0].y;
  y2=((pt.x-corners[2].x)*(corners[2].y-corners[3].y))/(corners[2].x-corners[3].x) + corners[2].y;
  return (( pt.x >= x1 && pt.x <= x2) && ( pt.y >=y1 && pt.y <=y2));
}

// returns the precentage of points within the detected Quadrilateral
float is_mostly_inside(vector<Point2f> sceneCorners, vector<Point2f> points)
{
  bool corner_cons=(sceneCorners[0].y < sceneCorners[2].y || sceneCorners[0].x < sceneCorners[2].x)&&
  (sceneCorners[0].x < sceneCorners[1].x && sceneCorners[2].y > sceneCorners[1].y)&&
  (sceneCorners[0].y < sceneCorners[3].y && sceneCorners[2].x > sceneCorners[3].x);
  float precentage=0;
  float count=points.size();
  //This means that the corners don't span a proper Quadrilateral but a pointy 8 shaped one.
  if(!corner_cons)
  {
    precentage=-1;
  }
  bool is_inside;
  //Check for each point if it's inside the Quadrilateral.
  for(int i=0; i<points.size() && corner_cons; i++)
  {
    is_inside=point_is_inside(sceneCorners, points[i]);
    if(is_inside)
      precentage = precentage + 1/count;
  }
    
  /* Uncomment this if you want to check how good your templates match.
  if(precentage == -1)
    printf("Not recognizing given template in scene. \n");
  else if(precentage>=0 && precentage <= ACCURACY)
    printf("precentage is %f Not accurate enough, recognition failed.\n", precentage);
  else
    printf("-- Precentage : %f \n", precentage );
  */
  return precentage;
}

//This function tries to look for template identified by template_num in the scene
void object_recognizing(int template_num)
{
  //-- Step 1: Detect the keypoints using SIFT Detector
  int minHessian = 400,i=0;
  SiftFeatureDetector detector( minHessian );
  std::vector<KeyPoint> keypoints_object, keypoints_scene;
  detector.detect( templates[template_num], keypoints_object );
  detector.detect( scene, keypoints_scene );
  
  //-- Step 2: Calculate descriptors (feature vectors)
  SiftDescriptorExtractor extractor;
  Mat descriptors_object, descriptors_scene;
  extractor.compute( templates[template_num], keypoints_object, descriptors_object );
  extractor.compute( scene, keypoints_scene, descriptors_scene );
  
  //-- Step 3: Matching descriptor vectors using FLANN matcher
  FlannBasedMatcher matcher;
  std::vector< DMatch > matches;
  matcher.match( descriptors_object, descriptors_scene, matches );
  double max_dist = 0; double min_dist = 100;
  
  //-- Quick calculation of max and min distances between keypoints
  for(i = 0; i < descriptors_object.rows; i++ )
  { 
    double dist = matches[i].distance;
    if( dist < min_dist ) min_dist = dist;
    if( dist > max_dist ) max_dist = dist;
  }
  //-- Draw only "good" matches (i.e. whose distance is less than 3*min_dist )
  std::vector< DMatch > good_matches;
  for(i = 0; i < descriptors_object.rows; i++ )
  {
    if( matches[i].distance < 3*min_dist )
    {
      good_matches.push_back( matches[i]);
    }
  }
  drawMatches( templates[template_num], keypoints_object, scene, keypoints_scene,
	       good_matches, scene_matches[template_num], Scalar::all(-1), Scalar::all(-1),
	       vector<char>(), DrawMatchesFlags::NOT_DRAW_SINGLE_POINTS );
  
  //-- Find the object in the scene using the good points from earlier.
  std::vector<Point2f> obj;
  std::vector<Point2f> scene_good_points;
  
  for(i = 0; i < good_matches.size(); i++ )
  {
    //-- Get the keypoints from the good matches
    obj.push_back( keypoints_object[ good_matches[i].queryIdx ].pt );
    scene_good_points.push_back( keypoints_scene[ good_matches[i].trainIdx ].pt );
  }
  Mat H;
  if(obj.size()>=4 && scene_good_points.size()>=4)
  {
    //Finds where is obj in the scene
    H = findHomography( obj, scene_good_points, CV_RANSAC );
    //-- Get the corners from the image_1 ( the object to be "detected" )
    std::vector<Point2f> obj(4);
    obj[0] = cvPoint(0,0); 
    obj[1] = cvPoint( templates[template_num].cols, 0 );
    obj[2] = cvPoint( templates[template_num].cols, templates[template_num].rows ); 
    obj[3] = cvPoint( 0, templates[template_num].rows );
    perspectiveTransform( obj, results[template_num].corners, H);
    /* Uncomment this if you want to see the result corners in each iteration:
    printf("result corners:\np0(%f,%f)\np1(%f,%f)\np2(%f,%f)\np3(%f,%f)\n",results[template_num].corners[0].x,results[template_num].corners[0].y,results[template_num].corners[1].x,results[template_num].corners[1].y,results[template_num].corners[2].x,results[template_num].corners[2].y,results[template_num].corners[3].x,results[template_num].corners[3].y);
    *///comment here
    
    //-- Draw lines between the corners
    
    /* Uncomment this if you want to draw the lines between the corners and write the result image later.
    line( scene_matches[template_num], results[template_num].corners[0] + Point2f( templates[template_num].cols, 0), results[template_num].corners[1] + Point2f( templates[template_num].cols, 0), Scalar(0, 255, 0), 4 );
    line( scene_matches[template_num], results[template_num].corners[1] + Point2f( templates[template_num].cols, 0), results[template_num].corners[2] + Point2f( templates[template_num].cols, 0), Scalar( 0, 255, 0), 4 );
    line( scene_matches[template_num], results[template_num].corners[2] + Point2f( templates[template_num].cols, 0), results[template_num].corners[3] + Point2f( templates[template_num].cols, 0), Scalar( 0, 255, 0), 4 );
    line( scene_matches[template_num], results[template_num].corners[3] + Point2f( templates[template_num].cols, 0), results[template_num].corners[0] + Point2f( templates[template_num].cols, 0), Scalar( 0, 255, 0), 4 );
    *///comment here
    results[template_num].precentage = is_mostly_inside(results[template_num].corners, scene_good_points); 
  }
  else
  {
    /* Uncomment this for notification of whether not enough good matches in the scene were detected.
    printf("not enough points\n");
    */
    results[template_num].precentage = -1;
    results[template_num].corners.clear();
    results[template_num].corners.push_back(Point2f(0.f,0.f));
    results[template_num].corners.push_back(Point2f(0.f,0.f));
    results[template_num].corners.push_back(Point2f(0.f,0.f));
    results[template_num].corners.push_back(Point2f(0.f,0.f));
  }
}
void object_recognition()
{
  int j=0, max_index=-1;
  //Activate the object recognition process for every template
  for(j=0; j<TEMPLATE_QUANTITY; j++)
  {
    object_recognizing(j);
  }
  //Find most fitting template according to precentage
  for(j=0; j<TEMPLATE_QUANTITY; j++)
  {
    if(results[j].precentage>=ACCURACY && j>max_index)
    {
      if(max_index>-1)
      {
	if( results[j].precentage>=results[max_index].precentage)
	{
	  max_index=j;
	}
	else;
      }
      else
      {
	max_index=j;
      }
    }
  }
  
  //If a match was found, return it in result
  if(max_index>-1)
  {
    result.precentage=results[max_index].precentage;
    result.corners=results[max_index].corners;
    /* Uncomment this if you want to see the result image, if the object was recognized 
    vector<int> compression_params;
    compression_params.push_back(CV_IMWRITE_JPEG_QUALITY);
    compression_params.push_back(95);
    //-- Show detected matches
    bool ans = imwrite("src/object_recognition/object_recognition_result.jpg" , scene_matches[max_index], compression_params);
    if(ans)
    {
      printf("Object Recognition Result image written\n");
    }
    else
    {
      printf("Failed to write Object Recognition Result image\n");
    }
    *///comment here
  }
  
  //Otherwise, returns 4 points of (0,0)
  else
  {
    result.precentage=0;
    result.corners.clear();
    result.corners.push_back(Point2f(0.f, 0.f));
    result.corners.push_back(Point2f(0.f, 0.f));
    result.corners.push_back(Point2f(0.f, 0.f));
    result.corners.push_back(Point2f(0.f, 0.f));
  }
}